@@ -8,7 +8,6 @@
 #include <QtCore/QUrl>
 #include <QtGrpc/qabstractgrpcclient.h>
 #include <QtGrpc/qgrpccallreply.h>
-#include <QtGrpc/qgrpccredentials.h>
 #include <QtGrpc/qgrpcstream.h>
 #include <QtNetwork/QNetworkAccessManager>
 #include <QtNetwork/QNetworkReply>
@@ -26,7 +25,7 @@ using namespace Qt::StringLiterals;
 
 /*!
     \class QGrpcHttp2Channel
-    \inmodule QtGRPC
+    \inmodule QtGrpc
 
     \brief The QGrpcHttp2Channel class is an HTTP/2 implementation of
     QAbstractGrpcChannel interface.
@@ -86,6 +85,35 @@ constexpr char GrpcStatusHeader[] = "grpc-status";
 constexpr char GrpcStatusMessageHeader[] = "grpc-message";
 constexpr qsizetype GrpcMessageSizeHeaderSize = 5;
 
+static void addMetadataToRequest(QNetworkRequest *request, const QGrpcMetadata &channelMetadata,
+                                 const QGrpcMetadata &callMetadata)
+{
+    auto iterateMetadata = [&request](const auto &metadata) {
+        for (const auto &[key, value] : std::as_const(metadata)) {
+            request->setRawHeader(key, value);
+        }
+    };
+
+    iterateMetadata(channelMetadata);
+    iterateMetadata(callMetadata);
+}
+
+static QGrpcMetadata collectMetadata(QNetworkReply *networkReply)
+{
+    return QGrpcMetadata(networkReply->rawHeaderPairs().begin(),
+                         networkReply->rawHeaderPairs().end());
+}
+
+static std::optional<std::chrono::milliseconds> deadlineForCall(
+        const QGrpcChannelOptions &channelOptions, const QGrpcCallOptions &callOptions)
+{
+    if (callOptions.deadline())
+        return *callOptions.deadline();
+    if (channelOptions.deadline())
+        return *channelOptions.deadline();
+    return std::nullopt;
+}
+
 struct QGrpcHttp2ChannelPrivate
 {
     struct ExpectedData
@@ -94,9 +122,8 @@ struct QGrpcHttp2ChannelPrivate
         QByteArray container;
     };
 
-    QUrl url;
     QNetworkAccessManager nm;
-    std::unique_ptr<QAbstractGrpcCredentials> credentials;
+    QGrpcChannelOptions channelOptions;
 #if QT_CONFIG(ssl)
     QSslConfiguration sslConfig;
 #endif
@@ -104,9 +131,9 @@ struct QGrpcHttp2ChannelPrivate
     QObject lambdaContext;
 
     QNetworkReply *post(QLatin1StringView method, QLatin1StringView service, QByteArrayView args,
-                        bool stream = false)
+                        const QGrpcCallOptions &callOptions)
     {
-        QUrl callUrl = url;
+        QUrl callUrl = channelOptions.host();
         callUrl.setPath("/%1/%2"_L1.arg(service, method));
 
         qGrpcDebug() << "Service call url:" << callUrl;
@@ -118,8 +145,8 @@ struct QGrpcHttp2ChannelPrivate
 #if QT_CONFIG(ssl)
         request.setSslConfiguration(sslConfig);
 #endif
-        for (const auto &[key, value] : credentials->callCredentials().toStdMap())
-            request.setRawHeader(key, value.toString().toUtf8());
+
+        addMetadataToRequest(&request, channelOptions.metadata(), callOptions.metadata());
 
         request.setAttribute(QNetworkRequest::Http2DirectAttribute, true);
 
@@ -140,9 +167,8 @@ struct QGrpcHttp2ChannelPrivate
                              QGrpcHttp2ChannelPrivate::abortNetworkReply(networkReply);
                          });
 #endif
-        if (!stream) {
-            // TODO: Add configurable timeout logic
-            QTimer::singleShot(6000, networkReply, [networkReply] {
+        if (auto deadline = deadlineForCall(channelOptions, callOptions)) {
+            QTimer::singleShot(*deadline, networkReply, [networkReply] {
                 QGrpcHttp2ChannelPrivate::abortNetworkReply(networkReply);
             });
         }
@@ -175,22 +201,22 @@ struct QGrpcHttp2ChannelPrivate
         return networkReply->readAll().mid(GrpcMessageSizeHeaderSize);
     }
 
-    QGrpcHttp2ChannelPrivate(const QUrl &_url,
-                             std::unique_ptr<QAbstractGrpcCredentials> _credentials)
-        : url(_url), credentials(std::move(_credentials))
+    QGrpcHttp2ChannelPrivate(const QGrpcChannelOptions &options) : channelOptions(options)
     {
 #if QT_CONFIG(ssl)
-        if (url.scheme() == "https"_L1) {
+        if (channelOptions.host().scheme() == "https"_L1) {
             // HTTPS connection requested but not ssl configuration provided.
-            Q_ASSERT(credentials->channelCredentials().contains(SslConfigCredential));
-            sslConfig = credentials->channelCredentials()
-                                .value(SslConfigCredential)
-                                .value<QSslConfiguration>();
-        } else if (url.scheme().isEmpty()) {
-            url.setScheme("http"_L1);
+            Q_ASSERT(channelOptions.sslConfiguration());
+            sslConfig = *channelOptions.sslConfiguration();
+        } else if (channelOptions.host().scheme().isEmpty()) {
+            auto tmpHost = channelOptions.host();
+            tmpHost.setScheme("http"_L1);
+            channelOptions.withHost(tmpHost);
         }
 #else
-        url.setScheme("http"_L1);
+        auto tmpHost = channelOptions.host();
+        tmpHost.setScheme("http"_L1);
+        channelOptions.withHost(tmpHost);
 #endif
     }
 
@@ -202,13 +228,10 @@ struct QGrpcHttp2ChannelPrivate
 };
 
 /*!
-    QGrpcHttp2Channel constructs QGrpcHttp2Channel with \a url used to
-    establish channel connections and \a credentials.
+    QGrpcHttp2Channel constructs QGrpcHttp2Channel with \a options.
 */
-QGrpcHttp2Channel::QGrpcHttp2Channel(const QUrl &url,
-                                     std::unique_ptr<QAbstractGrpcCredentials> credentials)
-    : QAbstractGrpcChannel(),
-      dPtr(std::make_unique<QGrpcHttp2ChannelPrivate>(url, std::move(credentials)))
+QGrpcHttp2Channel::QGrpcHttp2Channel(const QGrpcChannelOptions &options)
+    : QAbstractGrpcChannel(), dPtr(std::make_unique<QGrpcHttp2ChannelPrivate>(options))
 {
 }
 
@@ -220,15 +243,17 @@ QGrpcHttp2Channel::~QGrpcHttp2Channel() = default;
 /*!
     Synchronously calls the RPC method and writes the result to the output parameter \a ret.
 
-    The RPC method name is concatenated from the \a method and \a service parameters
-    with the given \a args.
+    The RPC method name is constructed by concatenating the \a method
+    and \a service parameters and called with the \a args argument.
+    Uses \a options argument to set additional parameter for the call.
 */
 QGrpcStatus QGrpcHttp2Channel::call(QLatin1StringView method, QLatin1StringView service,
-                                    QByteArrayView args, QByteArray &ret)
+                                    QByteArrayView args, QByteArray &ret,
+                                    const QGrpcCallOptions &options)
 {
     QEventLoop loop;
 
-    QNetworkReply *networkReply = dPtr->post(method, service, args);
+    QNetworkReply *networkReply = dPtr->post(method, service, args, options);
     QObject::connect(networkReply, &QNetworkReply::finished, &loop, &QEventLoop::quit);
 
     // If reply was finished in same stack it doesn't make sense to start event loop
@@ -246,20 +271,21 @@ QGrpcStatus QGrpcHttp2Channel::call(QLatin1StringView method, QLatin1StringView
 /*!
     Asynchronously calls the RPC method.
 
-    The RPC method name is concatenated from the \a method and \a service parameters
-    with the given \a args.
+    The RPC method name is constructed by concatenating the \a method
+    and \a service parameters and called with the \a args argument.
+    Uses \a options argument to set additional parameter for the call.
     The method can emit QGrpcCallReply::finished() and QGrpcCallReply::errorOccurred()
     signals on a QGrpcCallReply returned object.
 */
-std::shared_ptr<QGrpcCallReply> QGrpcHttp2Channel::call(QAbstractGrpcClient *client,
-                                                        QLatin1StringView method,
+std::shared_ptr<QGrpcCallReply> QGrpcHttp2Channel::call(QLatin1StringView method,
                                                         QLatin1StringView service,
-                                                        QByteArrayView args)
+                                                        QByteArrayView args,
+                                                        const QGrpcCallOptions &options)
 {
-    std::shared_ptr<QGrpcCallReply> reply(new QGrpcCallReply(client),
+    std::shared_ptr<QGrpcCallReply> reply(new QGrpcCallReply(serializer()),
                                           [](QGrpcCallReply *reply) { reply->deleteLater(); });
 
-    QNetworkReply *networkReply = dPtr->post(method, service, args);
+    QNetworkReply *networkReply = dPtr->post(method, service, args, options);
 
     auto connection = std::make_shared<QMetaObject::Connection>();
     auto abortConnection = std::make_shared<QMetaObject::Connection>();
@@ -269,18 +295,19 @@ std::shared_ptr<QGrpcCallReply> QGrpcHttp2Channel::call(QAbstractGrpcClient *cli
             [reply, networkReply, connection, abortConnection] {
                 QGrpcStatus::StatusCode grpcStatus = QGrpcStatus::StatusCode::Unknown;
                 QByteArray data = QGrpcHttp2ChannelPrivate::processReply(networkReply, grpcStatus);
+                reply->setMetadata(collectMetadata(networkReply));
                 QObject::disconnect(*connection);
                 QObject::disconnect(*abortConnection);
 
                 qGrpcDebug() << "RECV:" << data;
                 if (QGrpcStatus::StatusCode::Ok == grpcStatus) {
                     reply->setData(data);
-                    reply->finished();
+                    emit reply->finished();
                 } else {
                     reply->setData({});
-                    reply->errorOccurred({ grpcStatus,
-                                           QLatin1StringView(networkReply->rawHeader(
-                                                   GrpcStatusMessageHeader)) });
+                    emit reply->errorOccurred({ grpcStatus,
+                                                QLatin1StringView(networkReply->rawHeader(
+                                                        GrpcStatusMessageHeader)) });
                 }
                 networkReply->deleteLater();
             });
@@ -299,24 +326,31 @@ std::shared_ptr<QGrpcCallReply> QGrpcHttp2Channel::call(QAbstractGrpcClient *cli
 }
 
 /*!
-    Starts a stream on \a grpcStream using QGrpcStream::method() and
-    \a service to get the name of the RPC method.
+    Creates and starts a stream to the RPC method.
 
-    Calls QGrpcStream::handler() when the stream receives data from the server.
+    The RPC method name is constructed by concatenating the \a method
+    and \a service parameters and called with the \a arg argument.
+    Returns a shared pointer to the QGrpcStream. Uses \a options argument
+    to set additional parameter for the stream.
+
+    Calls QGrpcStream::updateData() when the stream receives data from the server.
     The method may emit QGrpcStream::errorOccurred() when the stream has terminated with an error.
 */
-void QGrpcHttp2Channel::startStream(QGrpcStream *grpcStream, QLatin1StringView service)
+std::shared_ptr<QGrpcStream> QGrpcHttp2Channel::startStream(QLatin1StringView method,
+                                                            QLatin1StringView service,
+                                                            QByteArrayView arg,
+                                                            const QGrpcCallOptions &options)
 {
-    Q_ASSERT(grpcStream != nullptr);
-    QNetworkReply *networkReply = dPtr->post(grpcStream->method(), service, grpcStream->arg(),
-                                             true);
+    QNetworkReply *networkReply = dPtr->post(method, service, arg, options);
 
+    std::shared_ptr<QGrpcStream> grpcStream(new QGrpcStream(method, arg, serializer()));
     auto finishConnection = std::make_shared<QMetaObject::Connection>();
     auto abortConnection = std::make_shared<QMetaObject::Connection>();
     auto readConnection = std::make_shared<QMetaObject::Connection>();
 
     *readConnection = QObject::connect(
-            networkReply, &QNetworkReply::readyRead, grpcStream, [networkReply, grpcStream, this] {
+            networkReply, &QNetworkReply::readyRead, grpcStream.get(),
+            [networkReply, grpcStream, this] {
                 auto replyIt = dPtr->activeStreamReplies.find(networkReply);
 
                 const QByteArray data = networkReply->readAll();
@@ -329,7 +363,7 @@ void QGrpcHttp2Channel::startStream(QGrpcStream *grpcStream, QLatin1StringView s
                                  << "expectedDataSize:" << expectedDataSize;
 
                     if (expectedDataSize == 0) {
-                        grpcStream->handler(QByteArray());
+                        grpcStream->updateData(QByteArray());
                         return;
                     }
 
@@ -350,7 +384,8 @@ void QGrpcHttp2Channel::startStream(QGrpcStream *grpcStream, QLatin1StringView s
                     qGrpcDebug() << "Full data received:" << data.size()
                                  << "dataContainer:" << dataContainer.container.size()
                                  << "capacity:" << dataContainer.expectedSize;
-                    grpcStream->handler(dataContainer.container.mid(
+                    grpcStream->setMetadata(collectMetadata(networkReply));
+                    grpcStream->updateData(dataContainer.container.mid(
                             GrpcMessageSizeHeaderSize,
                             dataContainer.expectedSize - GrpcMessageSizeHeaderSize));
                     dataContainer.container.remove(0, dataContainer.expectedSize);
@@ -369,10 +404,11 @@ void QGrpcHttp2Channel::startStream(QGrpcStream *grpcStream, QLatin1StringView s
                 }
             });
 
+    std::weak_ptr<QGrpcStream> weakGrpcStream(grpcStream);
     *finishConnection = QObject::connect(
-            networkReply, &QNetworkReply::finished, grpcStream,
-            [grpcStream, service, networkReply, abortConnection, readConnection, finishConnection,
-             this]() {
+            networkReply, &QNetworkReply::finished, grpcStream.get(),
+            [weakGrpcStream, service, networkReply, abortConnection, readConnection,
+             finishConnection, this]() {
                 const QString errorString = networkReply->errorString();
                 const QNetworkReply::NetworkError networkError = networkReply->error();
                 QObject::disconnect(*readConnection);
@@ -381,13 +417,16 @@ void QGrpcHttp2Channel::startStream(QGrpcStream *grpcStream, QLatin1StringView s
                 dPtr->activeStreamReplies.erase(networkReply);
                 QGrpcHttp2ChannelPrivate::abortNetworkReply(networkReply);
                 networkReply->deleteLater();
+
+                auto grpcStream = weakGrpcStream.lock();
+                if (!grpcStream) {
+                    qGrpcWarning() << "Could not lock gRPC stream pointer.";
+                    return;
+                }
                 qGrpcWarning() << grpcStream->method() << "call" << service
                                << "stream finished:" << errorString;
+                grpcStream->setMetadata(collectMetadata(networkReply));
                 switch (networkError) {
-                case QNetworkReply::RemoteHostClosedError:
-                    qGrpcDebug() << "Remote server closed connection. Reconnect silently.";
-                    startStream(grpcStream, service);
-                    break;
                 case QNetworkReply::NoError: {
                     // Reply is closed without network error, but may contain an unhandled data
                     // TODO: processReply returns the data, that might need the processing. It's
@@ -395,25 +434,25 @@ void QGrpcHttp2Channel::startStream(QGrpcStream *grpcStream, QLatin1StringView s
                     QGrpcStatus::StatusCode grpcStatus;
                     QGrpcHttp2ChannelPrivate::processReply(networkReply, grpcStatus);
                     if (grpcStatus != QGrpcStatus::StatusCode::Ok) {
-                        grpcStream->errorOccurred(QGrpcStatus{
-                                grpcStatus,
-                                QLatin1StringView(
-                                        networkReply->rawHeader(GrpcStatusMessageHeader)) });
+                        emit grpcStream->errorOccurred(
+                                QGrpcStatus{ grpcStatus,
+                                             QLatin1StringView(networkReply->rawHeader(
+                                                     GrpcStatusMessageHeader)) });
                     }
-                    grpcStream->finished();
                     break;
                 }
                 default:
-                    grpcStream->errorOccurred(QGrpcStatus{
-                            StatusCodeMap.at(networkError),
-                            "%1 call %2 stream failed: %3"_L1.arg(service, grpcStream->method(),
-                                                                  errorString) });
+                    emit grpcStream->errorOccurred(
+                            QGrpcStatus{ StatusCodeMap.at(networkError),
+                                         "%1 call %2 stream failed: %3"_L1.arg(
+                                                 service, grpcStream->method(), errorString) });
                     break;
                 }
+                emit grpcStream->finished();
             });
 
     *abortConnection = QObject::connect(
-            grpcStream, &QGrpcStream::finished, networkReply,
+            grpcStream.get(), &QGrpcStream::finished, networkReply,
             [networkReply, finishConnection, abortConnection, readConnection] {
                 QObject::disconnect(*finishConnection);
                 QObject::disconnect(*readConnection);
@@ -422,6 +461,8 @@ void QGrpcHttp2Channel::startStream(QGrpcStream *grpcStream, QLatin1StringView s
                 QGrpcHttp2ChannelPrivate::abortNetworkReply(networkReply);
                 networkReply->deleteLater();
             });
+
+    return grpcStream;
 }
 
 /*!
